import argparse
import os
import plistlib
import paramiko.ssh_exception
import pyasn1.codec.der.decoder
import pyasn1.type.univ
import socket
import tempfile
from .pyimg4 import IMG4
from fabric import Connection
from rich import print
from rich.prompt import Prompt
from typing import Optional, Tuple, Union


def introduction():
    print("Welcome to Extraire")
    print(
        "This program will dump blobs from your [bold]jailbroken[/bold] iOS device and copy it to your computer."
    )
    print(
        "Depending on the blob, you might [bold]not[/bold] be able to use it without a bootROM exploit (e.g. checkm8)."
    )
    print(
        "Refer to https://ios.cfw.guide/saving-blobs/#saving-onboard-blobs for more information.",
        end="\n\n",
    )


def interactive_input(
    address: Optional[str] = None,
    password: Optional[str] = None,
    port: Optional[int] = None,
) -> Tuple[str, str, int]:
    if None in (address, password, port) or '' in (address, password, port):
        print(
            "Options given out in [bold blue](blue parentheses)[/bold blue] are default values and will be used if you haven't specified an option.",
            end="\n\n",
        )

    if not address:
        while True:
            _address = Prompt.ask("[bold]Enter the device's IP address[/bold]")
            if _address == "":
                print("Please enter an IP address.")
                continue
            else:
                break
        print("")

    if not password:
        print(
            "Please note that it is normal for the password to [bold]not[/bold] appear."
        )
        _password = Prompt.ask(
            "[bold]Enter the root user password[/bold]", password=True, default="alpine"
        )
        print("")

    if not port:
        _port = int(Prompt.ask("[bold]Enter the SSH port[/bold]", default="22"))

    return (address or _address, password or _password, port or _port)


def dump_raw_apticket(address: str, password: str, port: int) -> Union[IMG4, bool]:
    with tempfile.TemporaryDirectory() as tmpdir:
        rawdump = os.path.join(tmpdir, "dump.raw")
        try:
            with Connection(
                address,
                user="root",
                port=port,
                connect_kwargs={
                    "password": password,
                },
                connect_timeout=10.0,
            ) as c:
                c.run("dd if=/dev/disk1 of=dump.raw bs=256 count=$((0x4000))")
                c.get("dump.raw", rawdump)
        except paramiko.ssh_exception.NoValidConnectionsError:
            print("[red]Could not connect to device.[/red]")
            return False
        except paramiko.ssh_exception.AuthenticationException:
            print("[red]Could not authenticate with your device.[/red]")
            print("[red]Please check if:[/red]")
            print("[red] - You entered the correct password[/red]")
            print("[red] - You have not disabled root login[/red]")
            return False
        except paramiko.ssh_exception.SSHException:
            print("[red]An SSH2 error occured.[/red]")
            return False
        except socket.timeout:
            print("[red]Timed out trying to connect.[/red]")
            return False
        except OSError as oserr:
            print(f"[red]OSError {oserr.errno} occured: {os.strerror(oserr.errno)}")
            return False
        except ValueError:
            print(
                f"[red]Please specify the port separately. {address} is not a valid input.[/red]"
            )
            return False

        with open(rawdump, "rb") as f:
            img4, _ = pyasn1.codec.der.decoder.decode(f.read())
    return IMG4(img4)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "host_port", metavar="HOST[:PORT]", nargs="?", help="The device's IP address", default=""
    )
    parser.add_argument(
        "-p", "--password", help="The device's root user password", required=False
    )
    parser.add_argument(
        "-o", "--output", help="Where to save the dumped blob", required=False
    )
    parser.add_argument(
        "--non-interactive",
        action="store_true",
        help="Don't interactively ask for missing value (assume default if missing)",
        required=False,
    )
    args = parser.parse_args()

    [address, port] = (
        args.host_port.split(":", 2)
        if ":" in args.host_port
        else [args.host_port, 22 if args.non_interactive else None]
    )

    introduction()
    if args.non_interactive:
        if not address:
            print("[red]Device address not specified, and user asked for non-interactive mode. Exiting.[/red]")
            return 1
        password = args.password or "alpine"
    else:
        address, password, port = interactive_input(address, args.password, port)

    img4 = dump_raw_apticket(address, password, port)
    if not img4:
        return 1

    ecid = str(img4.im4m["ECID"])
    args.output = f"{ecid}.blob.shsh2" if args.output is None else args.output
    with open(args.output, "wb") as f:
        plistlib.dump(img4.to_shsh(), f)

    print(f"[green]Done! Your blob has been saved to {args.output}[/green]")
    if 0x8020 <= int(img4.im4m["CHIP"]) < 0x8900:
        print(
            "[yellow][bold]Note:[/bold] Your device is probably an A12+ device.[/yellow]"
        )
        print(
            "[yellow]If you updated to your current version using the Settings app over-the-air, you [bold]cannot[/bold] use this blob, even with a jailbreak.[/yellow]"
        )
        print(
            "[yellow]Refer to https://ios.cfw.guide/saving-blobs/#ota-onboard-blobs for more information.[/yellow]"
        )
        print(
            "[yellow]Determine your blob type with https://verify.shsh.host or https://tsssaver.1conan.com/check or img4tool.[/yellow]"
        )
    return 0



import pyasn1.codec.der.decoder
import pyasn1.codec.der.encoder
import pyasn1.type.univ
from typing import Union, Any


__all__ = ["IMG4", "IM4R", "IM4M"]


def __remove_prefix(text: str, prefix: str) -> str:
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def endian_converter(val: Union[bytes, str, pyasn1.type.univ.OctetString]) -> str:
    """Converts big endian to small endian and vice versa
    Args:
        val (Union[bytes, str]): Bytes or a hex string to convert
    Returns:
        str: A 0x prefixed hex string
    """
    ba = bytearray()
    if isinstance(val, str):
        ba = bytearray.fromhex(__remove_prefix(val, "0x"))
    else:
        ba = bytearray(val)
    ba.reverse()
    s = "".join(format(x, "02x") for x in ba)
    return f"0x{s}"


class IM4R:
    def __init__(self, im4r: pyasn1.type.univ.Sequence):
        self.im4r = im4r

    @property
    def bncn(self) -> pyasn1.type.univ.Sequence:
        bncn = self.im4r[-1][-1]
        assert str(bncn[0]) == "BNCN"
        assert isinstance(bncn, pyasn1.type.univ.Sequence)
        return bncn


class IM4M:
    def __init__(self, im4m: pyasn1.type.univ.Sequence):
        self.im4m = im4m

    def __getitem__(self, key: str) -> Any:
        assert str(self.im4m[0]) == "IM4M"
        assert isinstance(self.im4m, pyasn1.type.univ.Sequence)

        set = self.im4m[2]
        manbpriv = set[0]
        assert str(manbpriv[0]) == "MANB"

        manbset = manbpriv[1]
        manppriv = manbset[0]
        assert str(manppriv[0]) == "MANP"

        manp = manppriv[1]

        for i in range(len(manp)):
            if str(manp[i][0]) == key:
                return manp[i][1]

        raise KeyError


class IMG4:
    def __init__(self, img4: pyasn1.type.univ.Sequence):
        self.img4 = img4

    @property
    def im4m(self) -> pyasn1.type.univ.Sequence:
        im4m = self.img4[2]
        assert str(im4m[0]) == "IM4M"
        assert isinstance(im4m, pyasn1.type.univ.Sequence)
        return IM4M(im4m)

    @property
    def im4r(self) -> pyasn1.type.univ.Sequence:
        im4r = self.img4[3]
        assert str(im4r[0]) == "IM4R"
        assert isinstance(im4r, pyasn1.type.univ.Sequence)
        return IM4R(im4r)

    def to_shsh(self) -> dict:
        generator = endian_converter(self.im4r.bncn[-1])

        _im4m = self.im4m.im4m
        _im4m.tagSet._TagSet__superTags = (_im4m.tagSet._TagSet__superTags[0],)

        _shsh = {}
        _shsh["ApImg4Ticket"] = pyasn1.codec.der.encoder.encode(_im4m)
        _shsh["generator"] = generator

        return _shsh



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['extraire']

package_data = \
{'': ['*']}

install_requires = \
['fabric>=2.7.0,<3.0.0', 'pyasn1>=0.4.8,<0.5.0', 'rich>=12.0.1,<13.0.0']

entry_points = \
{'console_scripts': ['extraire = extraire.__main__:main']}

setup_kwargs = {
    'name': 'extraire',
    'version': '0.1.4.post1',
    'description': 'Dumps onboard SHSH blobs with a valid generator for jailbroken iOS devices',
    'long_description': '<h1 align="center">extraire</h1>\n\n<p align="center">\n<a href="https://github.com/beerpiss/extraire/actions"><img alt="Actions Status" src="https://github.com/beerpiss/extraire/actions/workflows/build.yaml/badge.svg"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n<a href="https://github.com/beerpiss/extraire/blob/trunk/LICENSE"><img alt="License: 0BSD" src="https://img.shields.io/static/v1?label=License&message=0BSD&color=brightgreen"></a>\n<img alt="Supported Python versions: 3.6.2, 3.7, 3.8, 3.9, 3.10" src="https://img.shields.io/badge/python-3.6.2%20%7C%203.7%20%7C%203.8%20%7C%203.9%20%7C%203.10-blue">\n</p>\n\n> extraire (verb): to extract\n\nSimple program to dump onboard SHSH blobs with a valid generator for **jailbroken** iOS devices. Supports Windows, macOS and Linux.\n\n## What\'s this?\nThis program dumps the IMG4 ApTicket from /dev/disk1 on the device, copies it to your computer and converts it to a valid SHSH blob, no external dependencies required.\n\nEven though the dumped SHSH blob is valid, you will still be limited by a few factors:\n- SEP/Baseband/Rose firmware compatibility with the currently signed iOS version\n- If you\'ve updated to your current iOS version with the Settings app, you cannot use the dumped blob without a bootROM exploit (e.g. checkm8).\n\n## Requirements\nOpenSSH Server installed on your jailbroken device. That\'s it!\n\n## Installation\n### From PyPI\n```\npip install -U extraire\n```\n### Standalone binaries\nStandalone binaries for Windows, macOS and Linux can be found [here.](https://github.com/beerpiss/extraire/releases/tag/v0.1.4)\n\nYou will need to allow executable permission for macOS and Linux after downloading. Run `chmod +x /path/to/extraire` in a terminal (replace `/path/to/extraire` with the actual path).\n\n## Usage\nRun `extraire` only for an interactive guide.\n\n```\n❯ extraire --help\nusage: extraire [-h] [-p PASSWORD] [-o OUTPUT] [--non-interactive] [HOST[:PORT]]\n\npositional arguments:\n  HOST[:PORT]           The device\'s IP address\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -p PASSWORD, --password PASSWORD\n                        The device\'s root user password\n  -o OUTPUT, --output OUTPUT\n                        Where to save the dumped blob\n  --non-interactive     Don\'t interactively ask for missing value (assume default if missing)\n```\n\n## Running/building from source\nClone this repo, install the dependencies with `poetry install` or `pip install .`, and run `python3 -m extraire`\n\nTo build a wheel, do `poetry build`.\n\n## Credits\n[tihmstar](https://github.com/tihmstar): without his [img4tool](https://github.com/tihmstar/img4tool) code I wouldn\'t be able to write code for dealing with IMG4s in Python.\n',
    'author': 'beerpiss',
    'author_email': 'lacvtg.a1.2023@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<3.11',
}


setup(**setup_kwargs)



from sys import exit
from rich import print
from extraire.extraire import main


if __name__ == "__main__":
    try:
        exit(main())
    except KeyboardInterrupt:
        print("\n[red]Aborted by user.[/red]")
        exit(1)
